classdef BalchenModel < handle

    %% Initialize properties
    properties
        D;
        M;
        Neu;
        X;
        A_F;
        A_L;
        L;
    end

    methods

        %% Constructor
        function obj = BalchenModel(d,m,x,neu, A_F, A_L, L) 
            obj.D = d; %Drag Coefficients
            obj.M = m; % Momentum coefficients
            obj.X = x; % Initial States
            obj.Neu = neu;
            obj.A_F = A_F; %Frontal windage area
            obj.A_L = A_L; %Lateral windage area
            obj.L = L; %Vessel's length
        end

        %% Method that will be called to update the state 
        function [x, Y] = UpdateState(obj, u, W_c, W_v,gamma, dt)
            %W_c = Water current velocities in the NED frame
            %W_v = wind velocities in the NED frame
            %gamma = wind direction [deg] in NED frame

            V_c = obj.CurrentVelocity(W_c, obj.X);
            F_w = obj.WindForce(W_v, gamma, obj.X(3),obj.X(4), obj.X(5), obj.A_F, obj.A_L, obj.L);
            new_state = obj.RungeKutta(obj.X, u, F_w, V_c, dt, obj.M, obj.D, obj.Neu);
            obj.X = new_state;
            x = obj.X;
            Y = ConvertToNED(obj, obj.X(1:3));

        end

        function X = UpdateStatedlq(obj, u, W_c, W_v,gamma, dt)
            %W_c = Water current velocities in the NED frame
            %W_v = wind velocities in the NED frame
            %gamma = wind direction [deg] in NED frame

            V_c = obj.CurrentVelocity(W_c, obj.X);
            F_w = obj.WindForce(W_v, gamma, obj.X(3),obj.X(4), obj.X(5), obj.A_F, obj.A_L, obj.L);
            new_state = obj.RungeKutta(obj.X, u, F_w, V_c, dt, obj.M, obj.D, obj.Neu);
            obj.X = new_state;
            X = obj.X;
            X(1:3) = ConvertToNED(obj, X(1:3));
            X(4:6) = ConvertToNED(obj, X(4:6));
            
        end
        
        %% Method that contains the vessel model, it will be called by Runge-Kutta
        function f = VesselModel(obj, x, u, F_w, V_c, m, d, neu)

            f=zeros(6,1);
            f(1) = x(4);
            f(2) = x(5);
            f(3) = x(6);
            
            f(4) = - (d(1)/m(1))*abs(x(4) - V_c(1))*(x(4) - V_c(1)) + (1/m(1))*(F_w(1) + u(1)) + neu(1);
            f(5) = - (d(2)/m(2))*abs(x(5) - V_c(2))*(x(5) - V_c(2)) + (1/m(2))*(F_w(2) + u(2)) + neu(2);
            f(6) = - (d(3)/m(3))*abs(x(6))*x(6) - (d(4)/m(3))*abs(x(5) - V_c(2))*(x(5) - V_c(2)) + (1/m(3))*(F_w(3) + u(3) + V_c(3)) + neu(3);
        
        end
        
        %% Runge Kutta to integrate and get updated state values
        function x_next = RungeKutta(obj, x, u, F_w, v, dt, m, d, neu)

            K1 = obj.VesselModel(x, u, F_w, v, m, d, neu);
            K2 = obj.VesselModel(x+K1.*(dt/2), u, F_w, v, m, d, neu);
            K3 = obj.VesselModel(x+K2.*(dt/2), u, F_w, v, m, d, neu);
            K4 = obj.VesselModel(x+K3.*(dt), u, F_w, v, m, d, neu);
            
            x_next = x + (dt/6).*(K1+2.*K2+2.*K3 + K4);

        end

        %% Transform the water velocity to the Ship reference
        function V_c = CurrentVelocity(obj, W_c, state)  %Receives as input the current velocities and converts them to the ship frame of reference.

            R = [cos(state(3)) -sin(state(3)) 0;
                sin(state(3)) cos(state(3)) 0;
                0 0 1];
            
            V_c = R.'*W_c;
        
        end

        %% Calculate wind forces in the ship frame of reference
        function forcesArray = WindForce(obj, windVel, wAngle, vYaw,vSpeedSurge, vSpeedSway, A_F, A_L, L)
            %windVel: wind speed [m/s] in NED frame
            %wAngle: wind direction [deg] in NED frame
            %vYaw: vessel heading [rad]
            Cx = 0.6;
            Cy = 0.8;
            Cn = 0.1;
            % Cx, Cy, Cn: wind coefficients: assumed to be constants
            wAngle = wAngle * pi/180; 
            uw = windVel * cos(wAngle - vYaw); % windspeed in surge
            vw = windVel * sin (wAngle - vYaw);% windspeed in sway
            urw = vSpeedSurge - uw; % wind relative speed in surge
            vrw = vSpeedSway - vw; % wind relative speed in sway
            Vr = sqrt(urw^2 + vrw^2);%wind relative speed
            rho = 1.23; % wind density [kg\m^3]
            windForceSurge = 0.5*rho*Cx*A_F*cos(wAngle)* Vr^2; %Wind force in surge
            windForceSway = 0.5*rho*Cy*A_L*sin(wAngle)* Vr^2; %wind force in sway
            windMoment = 0.5*rho*Cn*A_L*L*sin(2*wAngle)* Vr^2; %Wind moment
            forcesArray = [windForceSurge, windForceSway, windMoment]';
        end

        %% Convert the states to NED
        function x_NED = ConvertToNED(obj, x)
            R = [cos(x(3)) -sin(x(3)) 0; sin(x(3)) cos(x(3)) 0; 0 0 1];
            x_NED = R*x;
            x_NED(3)=x_NED(3)*180/pi;
        end

        %% Convert from NED to ship frame
        function x_ship = ConvertFromNED(obj, x_NED)
            R = [cos(obj.X(3)) -sin(obj.X(3)) 0; sin(obj.X(3)) cos(obj.X(3)) 0; 0 0 1];
            x_ship = R'*x_NED;
            x_ship(3)=x_ship(3)*180/pi;
        end


    end

end
